<?php                     
	ini_set('display_startup_errors', 1);
	error_reporting(E_ALL);
class Database {
  public $dbconn;
  public $host;
  public $port;
  public $database;   
  public $user;
  public $password;
  
   function __construct($host, $port, $database, $user, $password) {
       $this->host = $host;   
       $this->port = $port;  
       $this->database = $database;
       $this->user = $user;
       $this->password = $password;
       $this->connect();
   }
   
    public function connect() {
       //Todo: port number
       $connStr = "host=" . $this->host . " dbname=" . $this->database . " user=" . $this->user . " password=" . $this->password . "";      
       $this->dbconn = pg_connect($connStr) or die(json_encode(array("success"=>false, "connStr"=>$connStr, "errMsg"=>'Could not connect: ' . pg_last_error())));         
	}        
	//Close the postgre conn 
    public function close() {
       //Todo: port number
	   if(get_resource_type($this->dbconn) == "pgsql link") {
		   try {
			   pg_close($this->dbconn);  
		   } catch (Exception $e) {}
	   }
       
    }
	//Upsert based on row duplication (primary field or constraints)
    public function UpsertOne($table, $fields, $options)  {
	    if($options == null) {
		  $options = array();
	    }
		$setClauses = array();
		$insertComponents = $this->getInsertComponents($fields);
		
		for($i = 0; $i < count($insertComponents["fieldList"]); $i++) {
			$insertComponents["fieldList"][$i] = '"' . $insertComponents["fieldList"][$i] . '"';
		}
		
		for($i = 0; $i < count($insertComponents["placeholders"]); $i++) {
			$setClauses[] = $insertComponents["fieldList"][$i] . "=" . $insertComponents["placeholders"][$i];
		}
		$userPlaceholder = -1;
		if(isset($options["userId"]) && ((isset($options["lastModifiedStats"]) && $options["lastModifiedStats"] == true) || (isset($options["createdStats"]) && $options["createdStats"] == true))) {
			//Add user id as last param
			$insertComponents["fieldValues"][] = $options["userId"];
			$userPlaceholder = '$' . (count($insertComponents["placeholders"]) + 1);
		}
		if(isset($options["createdStats"]) && $options["createdStats"] == true) {
			$insertComponents["fieldList"][] = '"_cBy"';
			$insertComponents["placeholders"][] = $userPlaceholder;
			$insertComponents["fieldList"][] = '"_cOn"';
			$insertComponents["placeholders"][] = 'CURRENT_TIMESTAMP';
		}                                 
		if(isset($options["lastModifiedStats"]) && $options["lastModifiedStats"] == true) {
			if(isset($options["userId"])) {
          $setClauses[] = '"_lmBy"=' . $userPlaceholder;
    			$insertComponents["fieldList"][] = '"_lmBy"';
    			$insertComponents["placeholders"][] = $userPlaceholder;
      }  
       
			$setClauses[] = '"_lm" = CURRENT_TIMESTAMP';
			$insertComponents["fieldList"][] = '"_lm"';
			$insertComponents["placeholders"][] = 'CURRENT_TIMESTAMP';
		}
		$duplicateClause = "";
		if(isset($options["duplicateField"])) {
			$duplicateClause = '("' . $options["duplicateField"] . '")';
		}
		if(isset($options["duplicateContraint"])) {
			$duplicateClause = 'ON CONSTRAINT "' . $options["duplicateContraint"] . '"';
		}
		$sqlQuery = 'INSERT INTO "' . $table . '" (' . implode(",", $insertComponents["fieldList"]) . ') VALUES (' . implode(",", $insertComponents["placeholders"]) . ') ON CONFLICT ' . $duplicateClause . ' DO UPDATE
						SET ' . implode(", ", $setClauses) . ';';
	//	exit($sqlQuery);
		return pg_query_params($sqlQuery, $insertComponents["fieldValues"]);
	}
	
    public function UpsertMulti($table, $records, $options)  {
		for($i = 0; $i < count($records); $i++) {
			$this->UpsertOne($table, $records[$i], $options);
		}
	}
	
    public function getInsertComponents($fieldDefs)  {
		$resp = array(
			"placeholders"=>array(),
			"fieldList"=>array(),
			"fieldValues"=>array()
		);
		for($i = 0; $i < count($fieldDefs); $i++) {
			if(isset($fieldDefs[$i]["type"]) && $fieldDefs[$i]["type"] == "ts_vector") {
				$resp["placeholders"][] = 'to_tsvector($' . ($i + 1) . ')';
			} else {
				//Standard case
				$resp["placeholders"][] = '$' . ($i + 1);
			}
			$resp["fieldList"][] = $fieldDefs[$i]["column"];
			$resp["fieldValues"][] = $fieldDefs[$i]["value"];
		}
		return $resp;
	}
    public function GetOneResult($query, $params, $fields, $options) {        
          if($options == null) {
            $options = array();
          }
         $multiTableFlag = $this->isMultiTableQuery($query);          
         $fieldList = $this->getFieldList($multiTableFlag, $fields);
         $query = $this->searchAndReplace($query, array("fieldList"=>$fieldList));  
         if(isset($options["whereQuery"])) {   
               $query .= $this->getWhereClauseFromQueryArray($options["whereQuery"], $fields); 
         }
         //exit("query:: " . $query);
         $result = pg_query_params($query, $params) or die('Query failed: ' . pg_last_error());
          if($resultObj = pg_fetch_array($result, null, PGSQL_ASSOC)) {
               return $this->prepResult($resultObj, $fields);
          }
          return false;
    }
    public function GetAllResults($query, $params, $fields, $options)  {
      if($options == null) {
        $options = array();
      }
     $results = array();
     $multiTableFlag = $this->isMultiTableQuery($query);
     $fieldList = $this->getFieldList($multiTableFlag, $fields);
     $replaceArray = array("fieldList"=>$fieldList);
     $query = $this->searchAndReplace($query, $replaceArray);
     if(isset($options["whereQuery"])) {
           $query .= $this->getWhereClauseFromQueryArray($options["whereQuery"], $fields); 
     }
     $result = pg_query_params($query, $params) or die('Query failed: ' . pg_last_error());
      while($resultObj = pg_fetch_array($result, null, PGSQL_ASSOC)) {
           $rowData = $this->prepResult($resultObj, $fields);
           if(isset($options["associative"]) && $options["associative"] === true) {
                $results[$rowData[$options["keyProp"]]] = $rowData;
           } else {
                $results[] = $rowData;
           }
      }
      return $results; 
    }
	
	
    public function GetAllResultsNested($query, $params, $nestingStructure) {
	  $fields = array();
	  $this->getFieldsFromNestingStructure($nestingStructure, $fields, array());
     $multiTableFlag = true;
     //$multiTableFlag = $this->isMultiTableQuery($query);
     $fieldList = $this->getFieldList($multiTableFlag, $fields);
     $replaceArray = array("fieldList"=>$fieldList);
     $query = $this->searchAndReplace($query, $replaceArray);
	  
	  
     $responseObj = array();
     $result = pg_query_params($query, $params) or die('Query failed: ' . pg_last_error());
      while($resultObj = pg_fetch_array($result, null, PGSQL_ASSOC)) {
           $rowData = $this->prepResult($resultObj, $fields);
           nestHelper($responseObj, $rowData, $nestingStructure, array());
      }
      return $responseObj; 
    }
	
	//Flattens the embedded fields
    public function getFieldsFromNestingStructure($nestingStructure, &$fields, $propPath) {
		if(isset($nestingStructure["keyProp"]) && isset($nestingStructure["Fields"])) {
			$keyField = $nestingStructure["keyProp"];
			$keyField["propPath"] =  $propPath;
			$keyField["propPath"][] = $keyField["prop"];
			$keyField["prop"] = implode("_", $keyField["propPath"]);
			$fields[] = $keyField;
			for($i = 0; $i < count($nestingStructure["Fields"]); $i++) {
				$fieldToAdd = $nestingStructure["Fields"][$i];
				$fieldToAdd["propPath"] =  $propPath;
				$fieldToAdd["propPath"][] = $fieldToAdd["prop"];
				$fieldToAdd["prop"] = implode("_", $fieldToAdd["propPath"]);
				$fields[] = $fieldToAdd;
			}
		}
		if(isset($nestingStructure["Embedded"])) {
			for($i = 0; $i < count($nestingStructure["Embedded"]); $i++) {
				$embedDef = $nestingStructure["Embedded"][$i];
				if(isset($embedDef["prop"]) && isset($embedDef["keyProp"]) && isset($embedDef["Fields"])) {
					$newPropPath = $propPath;
					$newPropPath[] = $embedDef["prop"];
					$this->getFieldsFromNestingStructure($embedDef, $fields, $newPropPath);
				}
			}
		}
	}
	
	
    public function DoesAssociationExist($mappingTable, $propDefs, $associationToCheck) {
		//Todo
	}
	
    public function GetAllAssociations($mappingTable, $propDefs, $matchRules) {
		//Todo
	}

	//options["updateType"] => "partial"(default), or ""
	//propDefs => array(  array("prop"=>"messageId"), array("prop"=>"to")  )
    public function EnsureAssociations($mappingTable, $propDefs, $associations, $exhaustive) {
		if(isset($exhaustive) && $exhaustive !== false) {
			//Remove others not in provided set
			$otherProp = -1;
			$myVal = -1;
			for($i = 0; $i < count($propDefs); $i++) {
				if($propDefs[$i]["prop"] != $exhaustive) {
					$otherProp = $propDefs[$i]["prop"];
				}
			}
			$safeList = array();
			if($otherProp != -1) {
				for($i = 0; $i < count($associations); $i++) {
					if(isset($associations[$i][$exhaustive])) {
						if($myVal == -1) {
							$myVal = $associations[$i][$exhaustive];
						} else if($myVal != $associations[$i][$exhaustive]) {
							return "There was a problem with your exhaustive request=> differing values for exhaustive pivot field)";
						}
					}
					if(isset($associations[$i][$otherProp])) {
						$oVal = $associations[$i][$otherProp];
						if(!in_array($oVal, $safeList)) {
							$safeList[] = $oVal;
						}
					}
				}
			}
			if(count($safeList) == 0) {
				$sqlQuery = 'DELETE FROM "' . $mappingTable . '" WHERE  "' . $exhaustive . '" = $1;';
				$repVals = array($myVal);
				pg_query_params($sqlQuery, $repVals);
			} else {
				$placeholders = array();
				$repVals = array($myVal);
				for($i = 0; $i < count($safeList); $i++) {
					$placeholders[] = '$' . ($i + 2);
					$repVals[] = $safeList[$i];
				}
				if($myVal == -1) {
					return "There was a problem with your exhaustive request=> differing values for exhaustive pivot field";
				}
				$sqlQuery = 'DELETE FROM "' . $mappingTable . '" WHERE "' . $exhaustive . '" = $1 AND "' . $otherProp . '" NOT IN (' . implode(", ", $placeholders) . ');';

				pg_query_params($sqlQuery, $repVals);
			}
		}
		$mapFields = array();
		for($i = 0; $i < count($propDefs); $i++) {
			$mapFields[] = '"' . $propDefs[$i]["prop"] . '"';
		}
		for($i = 0; $i < count($associations); $i++) {
			$newPropVals = getPropVals($associations[$i], $propDefs);
			$this->UpsertOne($mappingTable, $newPropVals, array(
				"duplicateField"=>implode(",", $mapFields)
			));
		}
	}

    public function RemoveAssociations($mappingTable, $propDefs, $associationsToDelete) {
		for($i = 0; $i < count($associationsToDelete); $i++) {
			$assoc = $associationsToDelete[$i];
			$deleteWhereClauses = array();
			$queryValues = array();
			
			$cnt = 0;
			foreach($assoc as $field=>$val) {
				$deleteWhereClauses[] = '"' . $field . '" = $' . ($cnt + 1);
				$cnt++;
				$queryValues[] = $val;
			}
			
			$sqlQuery = 'DELETE FROM "' . $mappingTable . '"  WHERE ' . implode(' and ', $deleteWhereClauses);
			pg_query_params($sqlQuery, $queryValues);
		}
		return array("success"=>true);
	}

    public function RemoveAllAssociations($mappingTable, $removeRules) {
		$deleteWhereClauses = array();
		$queryValues = array();
		if($removeRules != "all") {
			for($i = 0; $i < count($removeRules); $i++) {
				$deleteWhereClauses[] = '"' . $removeRules[$i]["field"] . '" = $' . ($i + 1);
				$queryValues[] = $removeRules[$i]["value"];
			}
		}
		$whereClause = "";
		if(count($deleteWhereClauses) > 0) {
			$whereClause = " WHERE " . implode(' and ', $deleteWhereClauses);
		}
		$sqlQuery = 'DELETE FROM "' . $mappingTable . '" ' . $whereClause;
		return pg_query_params($sqlQuery, $queryValues);
	}

    public function RunQuery($query) {
      //exit("Run this: " . $query);
      pg_query($query) or die(json_encode(array("success"=>false, "errMsg"=>'The query failed: ' . pg_last_error())));
    }
    public function RunQueryParams($query, $params) {
      //exit("Run this: " . json_encode($params));
      pg_query_params($query, $params) or die(json_encode(array("success"=>false, "errMsg"=>'The query failed: ' . pg_last_error())));
    }
    public function getWhereClauseFromQueryArray($query, $fields) {
         //Todo: expand this drastically
         $whereClause = ' WHERE ';
         $whereParts = array();
         $whereDelim = " AND ";
         foreach($query as $prop=>$val) {  
             if($prop == "_CId") {
                    $whereParts[] = '"' . $prop . '" = \'' . $val . "'";
             } else if($prop == "_iid") {
                    $whereParts[] = '"' . $prop . '" = \'' . $val . "'";
             } else {
                 for($i = 0; $i < count($fields); $i++) {
                      $fieldDef = $fields[$i];
                      if($fieldDef["prop"] == $prop)  {
                          if($fieldDef["type"] == "text") {
                                $whereParts[] = '"' . $prop . '" = \'' . $val . "'";
                          }
                          //Todo: ints decimals locs..
                      }
                 }
             }
         }
         if(count($whereParts) > 0) {
             return $whereClause . implode($whereDelim, $whereParts);
         } else {
            return "";
         }
    }
    
    public function getFieldList($multiTableFlag, $fields)  {
        $fieldListParts = array();
        for($i = 0; $i < count($fields); $i++) { 
              $colParts = explode(".", $fields[$i]["column"]);            
              if($multiTableFlag === true) { 
                  $fieldParts = "";
                  for($j = 0; $j < count($colParts); $j++) {
                    $fieldParts[] = '"' . $colParts[$j] . '"';
                  }
                  $fieldListParts[] = implode(".", $fieldParts) . ' as "' . $fields[$i]["prop"] . '"';
              }  else {
                  $colName = $colParts[count($colParts) - 1];
                  if($colName == $fields[$i]["prop"]) {
                       $fieldListParts[] = '"' . $fields[$i]["prop"] . '"';
                  } else {
                      $fieldListParts[] = '"' . $colName . '" as "' . $fields[$i]["prop"] . '"';
                  }
              }
            
        }
        return implode(", ", $fieldListParts);
    } 
    public function prepResult($dataObj, $fields) {
          $prepped = array();
          for($i = 0; $i < count($fields); $i++) { 
                $dataAtField = $dataObj[$fields[$i]["prop"]];
               if($fields[$i]["type"] == "json" || $fields[$i]["type"] == "freeformData" || $fields[$i]["type"] == "jsonb") {
               echo "prop: " . $fields[$i]["prop"] . "!!! daf: " . $dataAtField ;
                    if($dataAtField != null) {
                        $dataAtField = json_decode($dataAtField, true);
                    }   
					if(isset($fields[$i]["arrayType"]) && $fields[$i]["arrayType"] == "associative" && empty($dataAtField)) {
						$dataAtField = array("empty"=>1);
					}                      
               } else if($fields[$i]["type"] == "int") {
                    if($dataAtField != null && is_int($dataAtField)) {
                        $dataAtField = intval($dataAtField);
                    }  else if($dataAtField != null && is_numeric($dataAtField)) {
                        $dataAtField = (int)round(floatval($dataAtField));
                    }
               } else if($fields[$i]["type"] == "objectRef") {
                    //exit("objectRef: " . json_encode($dataAtField));
                    //$dataAtField = intval($dataAtField);
               }
                $prepped[$fields[$i]["prop"]] = $dataAtField;
          }              
          return $prepped;
    }
    public function searchAndReplace($search, $replace){
  		preg_match_all("/\{(.+?)\}/", $search, $matches);
  
  		if (isset($matches[1]) && count($matches[1]) > 0){
  			foreach ($matches[1] as $key => $value) {
  				if (array_key_exists($value, $replace)){
  					$search = preg_replace("/\{$value\}/", $replace[$value], $search);
  				}
  			}
  		}
  		return $search;
  	}
    public function getCurrentConnectDBName() {
         return pg_dbname($this->dbconn);
    }
	// Check if a table exists: Returns true or false
   public function DoesTableExist($tableName) {
      $this->connect();
        $existsQuery = "SELECT EXISTS (SELECT 1 FROM   information_schema.tables WHERE table_name = '" . $tableName . "');";
        //echo $existsQuery;
         $existsResult = pg_query($existsQuery) or die(json_encode(array("success"=>false, "errMsg"=>'Query failed: ' . pg_last_error())));
        if($existData = pg_fetch_array($existsResult, null, PGSQL_ASSOC)) {
			if($existData["exists"] == "t") {
                pg_free_result($existsResult);
                 return true;
            }
        }
        pg_free_result($existsResult);
        return false;
    }
       
    public function GetTableList() {
         $tableResults = $this->GetAllResults("SELECT {fieldList} FROM information_schema.tables WHERE table_schema <> 'pg_catalog' and table_schema <> 'information_schema';", array(), array(    
            array("prop"=>"table_name", "column"=>"COLUMNS.table_name", "type"=>"text")
         ), array());   
    		 $tables = array();
    		 for($i = 0; $i < count($tableResults); $i++) {
    			 $tables[] = $tableResults[$i]["table_name"];
    		 }
         return $tables;
         
    }    
    
    
    public function GetTableStructure($tableName) {
         $columnsStruct = $this->GetAllResults('SELECT {fieldList} from INFORMATION_SCHEMA.COLUMNS where table_name = $1;', array($tableName), array(    
            array("prop"=>"column_name", "column"=>"COLUMNS.column_name", "type"=>"text"),     
            array("prop"=>"data_type", "column"=>"COLUMNS.data_type", "type"=>"text"),
            array("prop"=>"is_nullable", "column"=>"COLUMNS.is_nullable", "type"=>"text"),
            array("prop"=>"numeric_precision", "column"=>"COLUMNS.numeric_precision", "type"=>"int"),
            array("prop"=>"numeric_scale", "column"=>"COLUMNS.numeric_scale", "type"=>"int"),
            array("prop"=>"dateTimePrec", "column"=>"COLUMNS.datetime_precision", "type"=>"int"),
            array("prop"=>"intervalType", "column"=>"COLUMNS.interval_type", "type"=>"text"),
            array("prop"=>"intervalPrec", "column"=>"COLUMNS.interval_precision", "type"=>"int"),
            array("prop"=>"maxLen", "column"=>"COLUMNS.character_maximum_length", "type"=>"int")
         ), array(
          "associative"=>true,
          "keyProp"=>"column_name"
         )); 
         return $columnsStruct;
    }
    public function GetAllTableStructures() {
         $tableResults = $this->GetAllResults("SELECT {fieldList} FROM information_schema.tables WHERE table_schema <> 'pg_catalog' and table_schema <> 'information_schema';", array(), array(    
            array("prop"=>"table_name", "column"=>"COLUMNS.table_name", "type"=>"text")
         ), array());   
    		 $tables = array();
    		 for($i = 0; $i < count($tableResults); $i++) {
    			 $tableName = $tableResults[$i]["table_name"];
           $tables[$tableName] = $this->GetTableStructure($tableName);
    		 }
         return $tables;
         
    }
    
    public function isMultiTableQuery($query) {
      if(stripos($query, " join ") !== false) {
        return true;
      }
      
      return false;
    } 
  }            
	// Check if a database exists: Returns true or false 
   function DoesDatabaseExist($databaseHost, $databasePort, $databaseName, $databaseUser, $databasePassword) {
       $connStr = "host=" . $databaseHost . " dbname=" . $databaseName . " user=" . $databaseUser . " password=" . $databasePassword . "";      
       $conn = pg_connect($connStr);
       if($conn == null || $conn == false) {
        return false;
       }
       pg_close($conn);
       return true; 
  }

	function getPropVals($dataObj, $relevantFields) {
		$propVals = array();
		for($i = 0; $i < count($relevantFields); $i++) {
			$fieldDef = $relevantFields[$i];
			if(isset($fieldDef["prop"])) {
				$type = "standard";
				if(isset($fieldDef["type"])) {
					$type = $fieldDef["type"];
				}
				if($type == "standard") {
					$prop = $fieldDef["prop"];
					if(isset($dataObj[$prop])) {
						$propVals[] = array("column"=>$prop, "value"=>$dataObj[$prop]);
					}
				} else if($type == "json") {
					$prop = $fieldDef["prop"];
					if(isset($dataObj[$prop])) {
						$propVals[] = array("column"=>$prop, "value"=>json_encode($dataObj[$prop]));
					}
				} else if($type == "textsearch") {
					$prop = $fieldDef["prop"];
					$searchFields = $fieldDef["searchFields"];
					$newValue = "";
					for($j = 0; $j < count($searchFields); $j++) {
						if(isset($dataObj[$searchFields[$j]])) {
							$newValue .= $dataObj[$searchFields[$j]] . " ";
						}
					}
					$propVals[] = array("column"=>$prop, "value"=>$newValue, "type"=>"ts_vector");
				}
			}
		}
		return $propVals;
	}
	
	
	//Recursive Function
	function nestHelper(&$mainData, $rowData, $nestingStructure, $curPath) {
		if(isset($nestingStructure["keyProp"])) {
			$keyPropFieldDef = $nestingStructure["keyProp"];
			$pathToKeyProp = $curPath;
			$pathToKeyProp[] = $keyPropFieldDef["prop"];
			$keyPropAlias = implode("_", $pathToKeyProp);
			$keyPropVal = $rowData[$keyPropAlias];
			if(!isset($mainData[$keyPropVal])) {
				$mainData[$keyPropVal] = array();
			}
			if(isset($nestingStructure["Fields"])) {
				for($i = 0; $i < count($nestingStructure["Fields"]); $i++) {
					$fieldDef = $nestingStructure["Fields"][$i];
					$fieldPath = $curPath;
					$fieldPath[] = $fieldDef["prop"];
					$fieldAlias = implode("_", $fieldPath);
					$mainData[$keyPropVal][$fieldDef["prop"]] = $rowData[$fieldAlias];
				}
			}
			if(isset($nestingStructure["Embedded"])) {
				$embeddedFieldDefs = $nestingStructure["Embedded"];
				for($i = 0; $i < count($embeddedFieldDefs); $i++) {
					$embedObjDef = $embeddedFieldDefs[$i];
					$newPath = $curPath;
					$newPath[] = $embedObjDef["prop"];
					if(!isset($mainData[$keyPropVal][$embedObjDef["prop"]])) {
						$mainData[$keyPropVal][$embedObjDef["prop"]] = array();
					}
					nestHelper($mainData[$keyPropVal][$embedObjDef["prop"]], $rowData, $embedObjDef, $newPath);
				}
				
			}
		}
	}
	
?>